import torch
import math
import numpy as np
from torch.distributions import Normal, kl_divergence
from scipy.sparse import csr_matrix
#from sklearn.neighbors import kneighbors_graph




def Negative_multinomial_loss(out,X,r):
    out+=1e-6
    device=out.device
    zero=torch.zeros((out.size(0), 1),device=device)
    if len(r)>1:
        vec_r=r.view(-1,1)
    else:
        vec_r=r*torch.ones((out.size(0), 1),device=device)
    out_cat=torch.cat((zero,out),dim=1)
    X_cat=torch.cat((vec_r,X),dim=1)
    log_p=torch.nn.LogSoftmax(dim=1)(out_cat+1e-8)
    loss=-torch.lgamma(torch.sum(X_cat,1)+1e-6) +torch.mean(torch.lgamma(vec_r+1e-6))- torch.sum(log_p*X_cat,1 )
    return loss.mean()/X.size(-1)

def multinomial_loss(out,X):
    out+=1e-6
    log_p=torch.nn.LogSoftmax(dim=1)(out)
    loss=-torch.sum(log_p*X,1 )
    return loss.mean()/X.size(-1)
    
    
    
def kl_div(mu, var):
    d=-0.5*(1.0+torch.log(var+1e-6)-mu**2-var)
    return torch.sum(d,dim=1).mean()

def kl_div_tst(mu, var):
    return kl_divergence(Normal(mu, var.sqrt()),
                         Normal(torch.zeros_like(mu),torch.ones_like(var))).sum(dim=1).mean()
 
def binary_cross_entropy(recon_x, x):
    return -torch.mean(x * torch.log(recon_x + 1e-8) + (1 - x) * torch.log(1 - recon_x + 1e-8), dim=-1) 



def clip_distance_matrix(pts_src: torch.Tensor, pts_dst: torch.Tensor, p: int = 2):
    """
    Returns the matrix of ||x_i-y_j||_p, while restricting x_i,y_j in unit balls by multiplying a constant.

    Parameters
    ----------
    pts_src
        [R, D] matrix
    pts_dst
        [C, D] matrix
    p
        p-norm
    
    Return
    ------
    [R, C] matrix
         clipped distance matrix
    """
    clip_value=torch.tensor(1000.0)
    max_norm=torch.max(  torch.abs(pts_src).max()+  torch.abs(pts_dst).max() , 2*clip_value )/2
    pts_src_clip=clip_value*pts_src/max_norm
    pts_dst_clip=clip_value*pts_dst/max_norm
    x_col = pts_src_clip.unsqueeze(1)
    y_row = pts_dst_clip.unsqueeze(0)
    distance = torch.sum((torch.abs(x_col - y_row)) ** p, 2)
    return distance








def unbalanced_ot_z( x1, x2,reg=0.1, reg_m=1.0,device='cpu',max_iteration={'outer':10,'inner':5}):
    '''
    Calculate a unbalanced optimal transport matrix between mini subsets.
    Parameters
    ----------
    reg:
        Entropy regularization parameter in OT
    reg_m:
        Unbalanced OT parameter. Larger values means more balanced OT
    device
        training device

    Returns
    -------
    matrix
        mini-subset unbalanced optimal transport matrix
    '''
    outer_iter=max_iteration['outer']
    inner_iter=max_iteration['inner']
    ns = x1.size(0)
    nt = x2.size(0)
    cost_pp = clip_distance_matrix(x1, x2, p=2)
    
    p_s = torch.ones(ns, 1) / ns   
    p_t = torch.ones(nt, 1) / nt
    

    p_s = p_s.to(device)
    p_t = p_t.to(device)

    tran = torch.ones(ns, nt) / (ns * nt)
    tran = tran.to(device)

    dual = (torch.ones(ns, 1) / ns).to(device)
    f = reg_m / (reg_m + reg)

    for m in range(outer_iter):
        cost = cost_pp
        kernel = torch.exp(-cost / (reg*torch.max(torch.abs(cost)))) * tran
        b = p_t / (torch.t(kernel) @ dual)
        for i in range(inner_iter):
            dual =( p_s / (kernel @ b) )**f
            b = ( p_t / (torch.t(kernel) @ dual) )**f
        tran = (dual @ torch.t(b)) * kernel
        
    out=tran.detach()
    if torch.isnan(out).sum() > 0:        
        out=None
    return  out



def Graph_Laplacian_torch(X,nearest_neighbor=10,t=1.0):
    XX=X.detach()
    D=clip_distance_matrix(XX,XX)
    values,indices=torch.topk(D, nearest_neighbor+1, dim=1,largest=False)
    pos=D>values[:,nearest_neighbor].view(-1,1)
    D[pos] = 0.0
    graph_data=D
    graph_data = graph_data + graph_data.T.multiply(graph_data.T > graph_data) - graph_data.multiply(graph_data.T > graph_data)
    W =graph_data
    index_pos = torch.where(W>0)
    W_mean=torch.mean(W[index_pos])
    W[index_pos] =torch.exp(-W[index_pos]/(t*W_mean))
    return (torch.diag(W.sum(1)) - W).detach()



def Transform(X,Y,T,L,lamda_Eigenvalue,eigenvalue_type):
    Y=Y.detach()
    T=T.detach()
    L=L.detach()
    
    if eigenvalue_type=='mean':
        a=T.sum(1)
        a_inv=1.0/a               
        lamda_Lapalcian=2*lamda_Eigenvalue/(torch.diag(L)*a_inv).mean() #diag(ALA).mean()
        l=2*torch.mm(T,Y)        
        M=lamda_Lapalcian*L+2*torch.diag(a)
        if torch.isnan(M).sum() > 0:
            print('Laplacian NAN')
            M_inv= torch.diag(a_inv)/2.0
        else:
            M_inv=torch.linalg.inv(M).to(torch.float32)
        result=torch.mm(M_inv,l)
    
    elif eigenvalue_type=='normal':
        lamda_Lapalcian=lamda_Eigenvalue
        l=2*torch.mm(T,Y)        
        M=lamda_Lapalcian*L+2*torch.diag(T.sum(1))
        M_inv=torch.linalg.inv(M).to(torch.float32)
        result=torch.mm(M_inv,l)
    return result





import torch
import torch.nn as nn
import torch.nn.functional as F
import scanpy as sc
import pandas as pd 
from scipy.sparse import csr_matrix
import numpy as np
import torch.autograd as autograd
import itertools
from tqdm import tqdm
from collections import defaultdict
from itertools import cycle
from torch.autograd import Variable
import random
import math
import sys
import time
import anndata as ad
from Fountain.layer import *
from Fountain.loss import *
from Fountain.utils import *
from Fountain.data import *


class Fountain(nn.Module):

    
    def __init__(self,adata, enc, dec, n_domain,r_negative_multinomial=1.0,batch_name='batch'):
        """
        Parameters
        ----------
        enc
            Encoder structure config
        dec
            Decoder structure config
        
        n_domain
            The number of batches
        """
        super().__init__()
        
        self.z_dim = enc[-1][1]
        
        n_batch=n_domain
        x_dim=dec[-1][1]
        self.n_cell=adata.shape[0]
        self.encoder = Encoder_vae(x_dim, enc)
        self.decoder = NN(self.z_dim+n_batch, dec)
        self.n_domain = n_domain       
        self.batch_ind=create_batchind_list(adata,batch_name)        
        self.r=nn.Parameter(torch.Tensor([r_negative_multinomial]),requires_grad=True) 
        del adata    
        
    def get_latent(self,dataloader,device='cuda:0'):
        self.to(device)
        n_cell=len(dataloader.dataset)
        X_latent=torch.zeros((n_cell, self.z_dim)).to(device)
        for x,domain_id,idx in dataloader:
            with torch.no_grad():
                X=x.float().to(device) 
                mu=self.encoder.forward(X)[1]
                X_latent[idx]=mu
        return  X_latent.cpu().numpy()
            
        
    
    def enhance(self,adata,device='cuda:0',latent_rep='mu',batch_name='batch'):
        X=torch.tensor(adata.X.todense()) .to(device) 
        self.to(device)
        z,mu,_=self.encoder(X)
        if latent_rep=="mu":
            rep=mu            
        elif latent_rep=="z":           
            rep=z
        batch_token=(0*torch.ones(z.size(0))).long().to(device)
        batch_token=F.one_hot(batch_token,num_classes=self.n_domain).to(torch.float32)
        recon_input=torch.cat((rep,batch_token),1)
        recon = self.decoder(recon_input)
        zero=torch.zeros((recon.size(0), 1),device=device)
        recon_cat=torch.cat((zero,recon),dim=1)
        P=torch.nn.Softmax(dim=1)(recon_cat+1e-10)
        p_0=P[:,0:1]    
        p=P[:,1:]
        r=((self.r)**2)*torch.ones_like(p_0)
        count=(r/p_0)*p       
        Enhance=torch.where(count > 1.0, torch.tensor(1), torch.tensor(0))
        enh=Enhance+X
        Enhance_out=torch.where( enh > 0.9, torch.tensor(1), torch.tensor(0))
        return csr_matrix(Enhance_out.detach().cpu().numpy()  , dtype=np.float32)

    
    def train(
            self,
            dataloader,             
            lambda_mse=0.005, 
            lambda_Eigenvalue=0.5,
            lambda_kl=0.5,
            lambda_recon=1.0, 
            reg=0.1,
            reg_m=5.0,
            lr_pre=1e-4,
            lr=2e-4,
            lr_r=5e-4,            
            max_iteration=30000,
            mid_iteration=3000,
            ot_iteration={'outer':50,'inner':2},
            knn_neighbors=10,
            eigenvalue_type='mean',
            loss='Negative_multinomial',            
            t=5.0,
            early_stopping=None,                        
            device='cuda:0',
            batch_name='batch',
            verbose=False,
        ):
       

        self.to(device)        
        lambda_L2=torch.tensor(0.0).to(device)  
        n_epoch = int(np.ceil(max_iteration/len(dataloader)))
        mid_epoch=int(np.ceil(mid_iteration/len(dataloader)))
        damp_iter=int(np.ceil(   (n_epoch-mid_epoch)/15    )   )
        if loss=='Negative_multinomial':
            loss_func=Negative_multinomial_loss
            optim = torch.optim.Adam([{'params': itertools.chain(self.encoder.parameters(), self.decoder.parameters()),'lr': lr_pre,'weight_decay':5e-4 }, {'params': self.r, 'lr': lr_r,'weight_decay':5e-4}])            
        elif loss=='multinomial':
            loss_func=multinomial_loss
            optim = torch.optim.Adam(self.parameters(), lr=lr_pre, weight_decay=5e-4)
        
             
        with tqdm(range(n_epoch), total=n_epoch, desc='Epochs') as tq:       
            for epoch in tq:
                
                if epoch>mid_epoch+1 and epoch<mid_epoch+damp_iter+2:
                    lambda_L2+=lambda_mse/damp_iter
                if epoch==mid_epoch+1:
                    
                    if loss=='Negative_multinomial' :
                        optim = torch.optim.Adam([{'params': itertools.chain(self.encoder.parameters(), self.decoder.parameters()),'lr': lr,'weight_decay':5e-4 }, {'params': self.r, 'lr': lr_r,'weight_decay':5e-4}])
                    
                    elif  loss=='multinomial':
                        optim = torch.optim.Adam(self.parameters(), lr=lr, weight_decay=5e-4)

                tk0 = tqdm(enumerate(dataloader), total=len(dataloader), leave=False, desc='Iterations', disable=(not verbose))
                epoch_loss = defaultdict(float)
                                                                                 
                for i, (x,domain_id,_) in tk0:
                    
                    recon_loss = torch.tensor(0.0).to(device)
                    kl_loss = torch.tensor(0.0).to(device)
                    mse_loss= torch.tensor(0.0).to(device)

                    x= x.float().to(device)
                    domain_id=domain_id.long().to(device) 
                    if len(torch.unique(domain_id)) < self.n_domain:
                        print('Omit Some Batches')
                    batch_dict={}                  
                    z_dict={}
                    z_dict_new={}
                    for j in range(self.n_domain):
                        loc_query = torch.where(domain_id==j)[0]                            
                        X_j=x[loc_query]
                        if X_j.size(0)==0:
                            continue
                        z_j, mu_j, var_j = self.encoder(X_j)

                        z_dict[j] = z_j
                        if j>0 and epoch>mid_epoch:
                            if z_j.size(0)<=5:
                                print('Too few samples')
                                continue
                            ot_Matz=unbalanced_ot_z(z_dict[0],z_dict[j],reg=reg,reg_m=reg_m,device=device, max_iteration=ot_iteration)
                            if ot_Matz==None:
                                print('None OT')
                                
                                continue
                            L_j=Graph_Laplacian_torch(z_j,nearest_neighbor=min(knn_neighbors,z_dict[j].size(0)-1),t=t)
                            z_dict_new[j]=Transform(z_j,z_dict[0],torch.t(ot_Matz),L_j,lambda_Eigenvalue,eigenvalue_type)

                            mse_loss += F.mse_loss( z_dict_new[j], z_j ) * X_j.size(-1) 


                        batch_token=(j*torch.ones(z_j.size(0))).long().to(device)
                        batch_token=F.one_hot(batch_token,num_classes=self.n_domain).to(torch.float32)
                        recon_input=torch.cat((z_j,batch_token),1)
                        recon_j = self.decoder(recon_input)

                        if loss=='Negative_multinomial': 
                            rr=(self.r)**2                                
                            recon_loss += loss_func( recon_j, X_j,rr ) * X_j.size(-1) 
                        else:
                            recon_loss += loss_func( recon_j, X_j ) * X_j.size(-1)
                        kl_loss += kl_div(mu_j, var_j)

                    loss_value ={'recon_loss':lambda_recon*recon_loss,'kl_loss':lambda_kl*kl_loss,'mse_loss':lambda_L2*mse_loss} 
                    optim.zero_grad()
                    loss_sum=sum(loss_value.values())                      
                    loss_sum.requires_grad_(True)
                    loss_sum.backward()
                    optim.step()
                    
                    for k,v in loss_value.items():
                        epoch_loss[k] += loss_value[k].item()

                    info = ','.join(['{}={:.3f}'.format(k, v) for k,v in loss_value.items()])
                    tk0.set_postfix_str(info)

                epoch_loss = {k:v/(i+1) for k, v in epoch_loss.items()}
                epoch_info = ','.join(['{}={:.3f}'.format(k, v) for k,v in epoch_loss.items()])
                tq.set_postfix_str(epoch_info) 

                if early_stopping is not None:
                    early_stopping(sum(epoch_loss.values()), self)

                    if early_stopping.early_stop:
                        print('EarlyStopping: run {} epoch'.format(epoch+1))
                        break   
      
                                                    


#!/usr/bin/env python

import numpy as np
import scanpy as sc
import matplotlib.pyplot as plt
import seaborn as sns
import anndata as  ad
            
            
def embedding(
        adata, 
        color='celltype', 
        color_map=None, 
        groupby='batch', 
        groups=None, 
        cond2=None, 
        v2=None, 
        save=None, 
        legend_loc='right margin', 
        legend_fontsize=None, 
        legend_fontweight='bold', 
        sep='_', 
        basis='X_umap',
        size=10,
        show=True,
    ):
    """
    plot separated embeddings with others as background
    
    Parameters
    ----------
    adata
        AnnData
    color
        meta information to be shown
    color_map
        specific color map
    groupby
        condition which is based-on to separate
    groups
        specific groups to be shown
    cond2
        another targeted condition
    v2
        another targeted values of another condition
    basis
        embeddings used to visualize, default is X_umap for UMAP
    size
        dot size on the embedding
    """
    
    if groups is None:
        groups = adata.obs[groupby].cat.categories
    
    
    for b in groups:
        adata.obs['tmp'] = adata.obs[color].astype(str)
        adata.obs['tmp'][adata.obs[groupby]!=b] = ''
        if cond2 is not None:
            adata.obs['tmp'][adata.obs[cond2]!=v2] = ''
            groups = list(adata[(adata.obs[groupby]==b) & 
                                (adata.obs[cond2]==v2)].obs[color].astype('category').cat.categories.values)
            size = min(size, 120000/len(adata[(adata.obs[groupby]==b) & (adata.obs[cond2]==v2)]))
        else:
            groups = list(adata[adata.obs[groupby]==b].obs[color].astype('category').cat.categories.values)
            size = min(size, 120000/len(adata[adata.obs[groupby]==b]))
        adata.obs['tmp'] = adata.obs['tmp'].astype('category')
        if color_map is not None:
            palette = [color_map[i] if i in color_map else 'gray' for i in adata.obs['tmp'].cat.categories]
        else:
            palette = None

        title = b if cond2 is None else v2+sep+b
        if save is not None:
            save_ = '_'+b+save
            show = False
        else:
            save_ = None
            show = True
        sc.pl.embedding(adata, color='tmp', basis=basis, groups=groups, title=title, palette=palette, size=size, save=save_,
                   legend_loc=legend_loc, legend_fontsize=legend_fontsize, legend_fontweight=legend_fontweight, show=show)
        del adata.obs['tmp']
        del adata.uns['tmp_colors']
        


        
        
        





import os
import numpy as np
import pandas as pd
import scipy
from scipy.sparse import issparse, csr
from scipy.sparse import csr_matrix
import torch
from torch.utils.data import Dataset
from torch.utils.data.sampler import Sampler
from torch.utils.data import DataLoader

from anndata import AnnData
import scanpy as sc
from sklearn.preprocessing import maxabs_scale, MaxAbsScaler

#from glob import glob

#np.warnings.filterwarnings('ignore')





class BatchSampler(Sampler):
    """
    Batch-specific Sampler
    sampled data of each batch is from the same dataset.
    """
    def __init__(self, batch_size, batch_id, drop_last=False):
        """
        create a BatchSampler object
        
        Parameters
        ----------
        batch_size
            batch size for each sampling
        batch_id
            batch id of all samples
        drop_last
            drop the last samples that not up to one batch
            
        """
        self.batch_size = batch_size
        self.drop_last = drop_last
        self.batch_id = batch_id

    def __iter__(self):
        batch = {}
        sampler = np.random.permutation(len(self.batch_id))
        for idx in sampler:
            c = self.batch_id[idx]
            if c not in batch:
                batch[c] = []
            batch[c].append(idx)

            if len(batch[c]) == self.batch_size:
                yield batch[c]
                batch[c] = []

        for c in batch.keys():
            if len(batch[c]) > 0 and not self.drop_last:
                yield batch[c]
            
    def __len__(self):
        if self.drop_last:
            return len(self.batch_id) // self.batch_size
        else:
            return (len(self.batch_id)+self.batch_size-1) // self.batch_size
    
def create_batchind_dict(adata,batch_name='batch'):#max
    max_batchid=adata.obs[batch_name].value_counts().idxmax()
    batchind_list=adata.obs[batch_name].cat.categories.tolist()
    max_batchindex=batchind_list.index(max_batchid)
    batchind_list[max_batchindex]=batchind_list[0]
    batchind_list[0]=max_batchid
    for j in range(len(batchind_list)):
        batch_ind_dict={batchind_list[i]:i for i in range(len(batchind_list))}
    return batch_ind_dict

def create_batchind_list(adata,batch_name='batch'):#max
    max_batchid=adata.obs[batch_name].value_counts().idxmax()
    batchind_list=adata.obs[batch_name].cat.categories.tolist()
    max_batchindex=batchind_list.index(max_batchid)
    batchind_list[max_batchindex]=batchind_list[0]
    batchind_list[0]=max_batchid
   
    return batchind_list
        


    
# def create_dataloader(adata,batch_size,batchind_dict,batch_name='batch',num_worker=4,droplast=False):
#     scdata = SingleCellDataset(adata,batchind_dict,batch_name) 
#     scdataloader = DataLoader(scdata, batch_size=batch_size,num_workers=num_worker,drop_last=droplast,shuffle=True)
#     return scdataloader
    
class SingleCellDataset(Dataset):
    """
    Dataloader of single-cell data
    """
    def __init__(self, adata, batchind_dict, batch_name='batch'):
        """
        Create a SingleCellDataset object
        
        Parameters
        ----------
        adata
            AnnData object wrapping the single-cell data matrix
        """
        self.adata = adata
        self.adata_batch = adata.obs[batch_name]
        self.batchind_dict = batchind_dict
        self.batch_name = batch_name
        
        

    def __len__(self):
        return self.adata.shape[0]
    
    def __getitem__(self, idx):
        
        x = torch.tensor(self.adata.X[idx].toarray().squeeze())

        # 获取 domain_id
        domain_id = self.adata_batch[idx]
        domain_id = self.batchind_dict[domain_id]
        
        return x, domain_id, idx

    

  
    
    
    
def create_dataloader(adata, batch_size, batchind_dict, batch_name='batch', num_worker=4, droplast=False):
    scdata = SingleCellDataset(adata, batchind_dict, batch_name) 
    scdataloader = DataLoader(scdata, batch_size=batch_size, num_workers=num_worker, drop_last=droplast, shuffle=True)
    return scdataloader










import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from sklearn.feature_extraction.text import TfidfTransformer
import scipy
import sklearn
from sklearn.neighbors import NearestNeighbors, KNeighborsRegressor
from sklearn.metrics import silhouette_score
from sklearn import metrics
from sklearn.preprocessing import OneHotEncoder, LabelEncoder,MaxAbsScaler
from sklearn.neighbors import NearestNeighbors, KNeighborsRegressor
def cluster_evaluation(adata_obs, label_key, cluster_key):
    '''
    Clustering Performance Evaluation

    Args:
        adata_obs: polars.internals.frame.DataFrame.
        label_key: e.g. 'cell type', 'cell_type'
        cluster_key: e.g. 'mc_Dleiden'

    Returns:
        evaluation

    '''
    print(cluster_key)
    AMI = sklearn.metrics.adjusted_mutual_info_score(adata_obs[label_key], adata_obs[cluster_key])
    ARI = sklearn.metrics.adjusted_rand_score(adata_obs[cluster_key], adata_obs[label_key])
    NMI = sklearn.metrics.normalized_mutual_info_score(adata_obs[cluster_key], adata_obs[label_key])
   
    return AMI,ARI,NMI









def overcorrection_score(emb, celltype, n_neighbors=100, n_pools=100, n_samples_per_pool=100, seed=124):
    n_neighbors = min(n_neighbors, len(emb) - 1)
    nne = NearestNeighbors(n_neighbors=1 + n_neighbors, n_jobs=8)
    nne.fit(emb)
    kmatrix = nne.kneighbors_graph(emb) - scipy.sparse.identity(emb.shape[0])

    score = 0
    celltype_ = np.unique(celltype)
    celltype_dict = celltype.value_counts().to_dict()
    
    N_celltype = len(celltype_)

    for t in range(n_pools):
        indices = np.random.choice(np.arange(emb.shape[0]), size=n_samples_per_pool, replace=False)
        score += np.mean([np.mean(celltype[kmatrix[i].nonzero()[1]][:min(celltype_dict[celltype[i]], n_neighbors)] == celltype[i]) for i in indices])

    return 1-score / float(n_pools)




def onehot(y, n):
    """
    Make the input tensor one hot tensors
    
    Parameters
    ----------
    y
        input tensors
    n
        number of classes
        
    Return
    ------
    Tensor
    """
    if (y is None) or (n<2):
        return None
    assert torch.max(y).item() < n
    y = y.view(y.size(0), 1)
    y_cat = torch.zeros(y.size(0), n).to(y.device)
    y_cat.scatter_(1, y.data, 1)
    return y_cat

def token_map(x,j,n_domain,device):
    #token is a one hot vector
    batch_token=(j*torch.ones(x.size(0))).to(torch.int64).to(device)
    batch_token=F.one_hot(batch_token,num_classes=n_domain).to(torch.float32)
    return torch.cat((x,batch_token),1)

def TFIDF(adata,tfidf):
    # Perform TF-IDF (count_mat: peak*cell)
    def tfidf1(count_mat): 
        nfreqs = 1.0 * count_mat / np.tile(np.sum(count_mat,axis=0), (count_mat.shape[0],1))
        tfidf_mat = np.multiply(nfreqs, np.tile(np.log(1 + 1.0 * count_mat.shape[1] / np.sum(count_mat,axis=1)).reshape(-1,1), (1,count_mat.shape[1])))
        return scipy.sparse.csr_matrix(tfidf_mat)

    # Perform Signac TF-IDF (count_mat: peak*cell)
    def tfidf2(count_mat): 
        tf_mat = 1.0 * count_mat / np.tile(np.sum(count_mat,axis=0), (count_mat.shape[0],1))
        signac_mat = np.log(1 + np.multiply(1e4*tf_mat,  np.tile((1.0 * count_mat.shape[1] / np.sum(count_mat,axis=1)).reshape(-1,1), (1,count_mat.shape[1]))))
        return scipy.sparse.csr_matrix(signac_mat)
 
    def tfidf3(count_mat): 
        model = TfidfTransformer(smooth_idf=False, norm="l2")
        model = model.fit(np.transpose(count_mat))
        model.idf_ -= 1
        tf_idf = np.transpose(model.transform(np.transpose(count_mat)))
        return scipy.sparse.csr_matrix(tf_idf)
    
    if tfidf=='tfidf0':
        X_norm = sc.pp.normalize_total(adata, inplace=False)['X']
        adata.X = X_norm.copy()
    if tfidf=='tfidf1': 
        tfidf_res = tfidf1(adata.X.T).T
        adata.X = tfidf_res.copy()
    if tfidf=='tfidf2': 
        tfidf_res = tfidf2(adata.X.T).T
        adata.X = tfidf_res.copy()
    if tfidf=='tfidf3': 
        tfidf_res = tfidf3(adata.X.T).T
        adata.X = tfidf_res.copy()
        
        
        
class EarlyStopping:
    """
    Early stops the training if loss doesn't improve after a given patience.
    """
    def __init__(self, patience=10, verbose=False, checkpoint_file=''):
        """
        Parameters
        ----------
        patience 
            How long to wait after last time loss improved. Default: 30
        verbose
            If True, prints a message for each loss improvement. Default: False
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.loss_min = np.Inf
        self.checkpoint_file = checkpoint_file

    def __call__(self, loss, model):
        if np.isnan(loss):
            self.early_stop = True
        score = -loss

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(loss, model)
        elif score <= self.best_score:
            self.counter += 1
            if self.verbose:
                print(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
                model.load_model(self.checkpoint_file)
        else:
            self.best_score = score
            self.save_checkpoint(loss, model)
            self.counter = 0

    def save_checkpoint(self, loss, model):
        '''
        Saves model when loss decrease.
        '''
        if self.verbose:
            print(f'Loss decreased ({self.loss_min:.6f} --> {loss:.6f}).  Saving model ...')
        torch.save(model.state_dict(), self.checkpoint_file)
        self.loss_min = loss
        
class EarlyStopping_simple:
    """
    Early stops the training if loss doesn't improve after a given patience.
    """
    def __init__(self, patience=10, verbose=False):
        """
        Parameters
        ----------
        patience 
            How long to wait after last time loss improved. Default: 30
        verbose
            If True, prints a message for each loss improvement. Default: False
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.loss_min = np.Inf
       

    def __call__(self, loss, model):
        if np.isnan(loss):
            self.early_stop = True
        score = -loss

        if self.best_score is None:
            self.best_score = score
            
        elif score <= self.best_score:
            self.counter += 1
            if self.verbose:
                print(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
                
        else:
            self.best_score = score
           
            self.counter = 0

    def save_checkpoint(self, loss, model):
        '''
        Saves model when loss decrease.
        '''
        if self.verbose:
            print(f'Loss decreased ({self.loss_min:.6f} --> {loss:.6f}).  Saving model ...')
      
        self.loss_min = loss   



# encoding=utf-8
from setuptools import setup, find_packages
import pathlib

#here = pathlib.Path(__file__).parent.resolve()

setup(
    name="scFountain",
    version="0.0.7",
    description="Rigorous integration of single-cell ATAC-seq data using regularized barycentric mapping",
    long_description='Single-cell assay for transposase-accessible chromatin using sequencing (scATAC-seq) deciphers genome-wide chromatin accessibility, providing profound insights into gene regulation mechanisms. With the rapid advance of sequencing technology, scATAC-seq data typically encompasses numerous samples generated under various conditions, leading to complex and multifactorial batch effects, thus necessitating reliable batch integration tools. Although numerous batch integration tools exist for single-cell RNA sequencing (scRNA-seq) data, their effectiveness on scATAC-seq data has proven limited due to the characteristic differences between scRNA-seq and scATAC-seq data. Existing integration methods for scATAC-seq data suffer from several fundamental limitations, such as disrupting the  biological heterogeneity and focusing solely on low-dimensional correction, which can lead to data distortion and hinder downstream analysis. Here we propose Fountain, a deep learning framework for scATAC-seq data integration via rigorous barycentric mapping. Fountain regularizes barycentric mapping with geometric data information to achieve biological heterogeneity-preserving integration. Through comprehensive experiments on multiple datasets involving various laboratory protocols, sample composition and species, we demonstrate the advantages of Fountain over existing methods in batch correction and biological conservation. Additionally, the trained Fountain model can integrate data from new batches alongside already integrated data without retraining, thereby facilitating the incorporation of additional data and enabling continuous online data integration. Moreover, we provide a reconstruction strategy to obtain batch-corrected ATAC profiles, which has been proven to  better capture cellular heterogeneity and reveal cell type-specific implications such as expression enrichment analysis and partitioned heritability analysis. ',
    license="MIT Licence",
    url="https://github.com/BioX-NKU/Fountain",
    author="BioX-NKU",
    classifiers=[
        "Programming Language :: Python :: 3.9",
    ],
    keywords="scATAC-seq, batch integration, barycentric mapping, online integration, cell type-specific implication",
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        'anndata>=0.8.0',
        'scanpy>=1.9.1',
        'torch>=1.12.1',
        'pandas>=1.4.2',
        'scikit-learn>=1.0.2',
        'numpy>=1.21.5',
        'scipy>=1.9.3',
        'episcanpy>=0.3.2',
        'tqdm>=4.28.1',      
    ]
)











import math
import numpy as np

import torch
from torch import nn as nn
import torch.nn.functional as F
import torch.autograd as autograd
from torch.distributions import Normal
from torch.nn.parameter import Parameter
from torch.nn import init
from torch.autograd import Function


activation = {
    'relu':nn.ReLU(),
    'rrelu':nn.RReLU(),
    'sigmoid':nn.Sigmoid(),
    'leaky_relu':nn.LeakyReLU(),
    'tanh':nn.Tanh(),
    'gelu':nn.GELU(),
    'softmax':nn.Softmax(dim=1),
    '':None
}


class DSBatchNorm(nn.Module):
    """
    Domain-specific Batch Normalization
    """
    def __init__(self, num_features, n_domain, eps=1e-5, momentum=0.1):
        """
        Parameters
        ----------
        num_features
            dimension of the features
        n_domain
            domain number
        """
        super().__init__()
        self.n_domain = n_domain
        self.num_features = num_features
        self.bns = nn.ModuleList([nn.BatchNorm1d(num_features, eps=eps, momentum=momentum) for i in range(n_domain)])
        
    def reset_running_stats(self):
        for bn in self.bns:
            bn.reset_running_stats()
            
    def reset_parameters(self):
        for bn in self.bns:
            bn.reset_parameters()
            
    def _check_input_dim(self, input):
        raise NotImplementedError
    #y int       
    def forward(self, x, y):
        out = torch.zeros(x.size(0), self.num_features, device=x.device) #, requires_grad=False)
        
        out = self.bns[y](x)
            
        return out
        

class Block(nn.Module):
    """
    Basic block consist of:
        fc -> bn -> act -> dropout
    """
    def __init__(
            self,
            input_dim, 
            output_dim, 
            norm='', 
            act='', 
            dropout=0
        ):
        """
        Parameters
        ----------
        input_dim
            dimension of input
        output_dim
            dimension of output
        norm
            batch normalization, 
                * '' represent no batch normalization
                * 1 represent regular batch normalization
                * int>1 represent domain-specific batch normalization of n domain
        act
            activation function,
                * relu -> nn.ReLU
                * rrelu -> nn.RReLU
                * sigmoid -> nn.Sigmoid()
                * leaky_relu -> nn.LeakyReLU()
                * tanh -> nn.Tanh()
                * '' -> None
        dropout
            dropout rate
        """
        super().__init__()
        self.fc = nn.Linear(input_dim, output_dim)
        nn.init.xavier_uniform_(self.fc.weight)
        if type(norm) == int:
            if norm==1: # TO DO
                self.norm = nn.BatchNorm1d(output_dim)
            else:
                self.norm = DSBatchNorm(output_dim, norm)
        else:
            self.norm = None
            
        self.act = activation[act]
            
        if dropout >0:
            self.dropout = nn.Dropout(dropout)
        else:
            self.dropout = None
            
    def forward(self, x, y=None):
        h = self.fc(x)
        if self.norm:
            if len(x) == 1:
                pass
            elif self.norm.__class__.__name__ == 'DSBatchNorm':
                h = self.norm(h, y)
            else:
                h = self.norm(h)
        if self.act:
            h = self.act(h)
        if self.dropout:
            h = self.dropout(h)
        return h
        
    
class NN(nn.Module):
    """
    Neural network consist of multi Blocks
    """
    def __init__(self, input_dim, cfg):
        """
        Parameters
        ----------
        input_dim
            input dimension
        cfg
            model structure configuration, 'fc' -> fully connected layer
            
        Example
        -------
        >>> latent_dim = 10
        >>> dec_cfg = [['fc', x_dim, n_domain, 'sigmoid']]
        >>> decoder = NN(latent_dim, dec_cfg)
        
        
        """
        super().__init__()
        net = []
        for i, layer in enumerate(cfg):
            if i==0:
                d_in = input_dim
            if layer[0] == 'fc':
                net.append(Block(d_in, *layer[1:]))
            d_in = layer[1]
        self.net = nn.ModuleList(net)
    
    def forward(self, x, y=None):
        for layer in self.net:
            x = layer(x,y)
        return x
    
        


class Encoder_vae(nn.Module):
    """
    VAE Encoder
    """
    def __init__(self, input_dim, cfg):
        """
        Parameters
        ----------
        input_dim
            input dimension
        cfg
            encoder configuration, e.g. enc_cfg = [['fc', 1024, 1, 'relu'],['fc', 10, '', '']]
        """
        super().__init__()
        h_dim = cfg[-2][1]
        self.enc = NN(input_dim, cfg[:-1])
        self.mu_enc = NN(h_dim, cfg[-1:])
        self.var_enc = NN(h_dim, cfg[-1:])
        
    def reparameterize(self, mu, var):
        return mu+var.sqrt()*torch.randn(mu.size(),device=mu.device)

    def forward(self, x, y=None):
        q = self.enc(x, y)
        mu = self.mu_enc(q, y)
        var = torch.exp(self.var_enc(q, y))+1e-6
        z = self.reparameterize(mu, var)
        return z, mu, var
                                             

    
    
 

        
        
        
   
        
       